"""Client for interacting with the ClickZetta API."""

from __future__ import absolute_import
from __future__ import division

import re
import time
import shutil
from urllib.parse import urlparse, unquote

import clickzetta
from sqlalchemy.engine.url import make_url
import json
import random
import os
import requests.exceptions
import typing
from typing import (
    Any,
    Dict,
    IO,
    Iterable,
    Mapping,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)
import uuid
import warnings
import logging
from google.protobuf.json_format import MessageToJson, Parse, ParseDict
from datetime import datetime

from clickzetta._helpers import _get_click_zetta_host, _DEFAULT_HOST

from clickzetta.enums import *
from clickzetta.table import Table
from clickzetta.query_result import QueryResult, QueryData, QueryDataType
from clickzetta.proto.generated import ingestion_pb2
from clickzetta.bulkload.bulkload_enums import BulkLoadOptions, BulkLoadMetaData, BulkLoadOperation, \
    BulkLoadCommitMode, BulkLoadConfig
from clickzetta.utils import split_sql
from clickzetta._volume import resolve_local_path

TIMEOUT_HEADER = "X-Server-Timeout"
TimeoutType = Union[float, None]
ResumableTimeoutType = Union[
    None, float, Tuple[float, float]
]

if typing.TYPE_CHECKING:
    PathType = Union[str, bytes, os.PathLike[str], os.PathLike[bytes]]
    import requests

_DEFAULT_CHUNKSIZE = 10 * 1024 * 1024  # 10 MB
_MAX_MULTIPART_SIZE = 5 * 1024 * 1024
_DEFAULT_NUM_RETRIES = 6
_GENERIC_CONTENT_TYPE = "*/*"

_MIN_GET_QUERY_RESULTS_TIMEOUT = 120
HTTP_PROTOCOL_DEFAULT_PORT = '80'
HTTPS_PROTOCOL_DEFAULT_PORT = '443'
HTTP_PROTOCOL_PREFIX = 'http://'
HTTPS_PROTOCOL_PREFIX = 'https://'

DEFAULT_TIMEOUT = None

FIRST_SQL_SUBMIT_TIMEOUT = 30
HYBRID_SQL_TIMEOUT = 120

HEADERS = {
    'Content-Type': 'application/json'
}

DEFAULT_INSTANCE_ID = 100

class Client(object):

    def __init__(self, username: str = None, password: str = None, instance: str = None, workspace: str = None,
                 vcluster: str = None, cz_url: str = None, service: str = None, schema: str = None,
                 magic_token: str = None, protocol: str = None, hints: {} = None,
                 **extra):
        self.schema = None
        self.magic_token = None
        self.protocol = None
        self.hints = hints
        if cz_url is not None:
            self._parse_url(cz_url)
            self.instance_id = 0
        else:
            if service is None:
                raise ValueError('ClickZetta connection url `service` is required.')
            else:
                if protocol is None or protocol == 'https':
                    self.service = HTTPS_PROTOCOL_PREFIX + service if not service.startswith(
                        HTTPS_PROTOCOL_PREFIX) else service
                elif protocol == 'http':
                    self.service = HTTP_PROTOCOL_PREFIX + service if not service.startswith(
                        HTTP_PROTOCOL_PREFIX) else service
                else:
                    raise ValueError('protocol must be http or https. Other protocols are not supported yet.')
            self.token = None
            self.workspace = workspace
            self.instance = instance
            self.vcluster = vcluster
            self.session = None
            self.instance_id = 0
            self.username = username
            self.schema = schema
            self.password = password
            self.magic_token = magic_token
            self.protocol = protocol
        if self.magic_token is not None:
            self.token = self.magic_token
        else:
            self.token = self.log_in_cz(self.username, self.password, self.instance)

    def _parse_url(self, url: string):
        url = make_url(url)
        query = dict(url.query)
        port = url.port

        self.instance = url.host.split('.')[0]
        length = len(self.instance) + 1

        if 'protocol' in query:
            self.protocol = query.pop('protocol')
            if self.protocol == 'http':
                if not port:
                    self.service = HTTP_PROTOCOL_PREFIX + url.host[length:] + ':' + HTTP_PROTOCOL_DEFAULT_PORT
                else:
                    self.service = HTTP_PROTOCOL_PREFIX + url.host[length:] + ':' + str(port)
            else:
                raise ValueError('protocol parameter must be http. Other protocols are not supported yet.')
        else:
            self.protocol = 'https'
            if not port:
                self.service = HTTPS_PROTOCOL_PREFIX + url.host[length:] + ':' + HTTPS_PROTOCOL_DEFAULT_PORT
            else:
                self.service = HTTPS_PROTOCOL_PREFIX + url.host[length:] + ':' + str(port)

        self.workspace = url.database
        self.username = url.username
        self.password = url.password

        if 'virtualcluster' in query or 'virtualCluster' in query or 'vcluster' in query:
            if 'virtualcluster' in query:
                self.vcluster = query.pop('virtualcluster')
            elif 'virtualCluster' in query:
                self.vcluster = query.pop('virtualCluster')
            else:
                self.vcluster = query.pop('vcluster')
        else:
            raise ValueError('url must have `virtualcluster` or `virtualCluster` or `vcluster` parameter.')
        if 'schema' in query:
            self.schema = query.pop('schema')
        if 'magic_token' in query:
            self.magic_token = query.pop('magic_token')

    def log_in_cz(self, username: str, password: str, instance: str) -> str:
        path = "/clickzetta-portal/user/loginSingle"
        login_params = LoginParams(username, password, instance)
        api_repr = login_params.to_api_repr()
        data = json.dumps(api_repr)
        try:
            api_response = requests.post(self.service + path, data=data, headers=HEADERS, timeout=10)
            if api_response.status_code != 200:
                raise requests.exceptions.RequestException(
                    "user:" + login_params.username + " login to clickzetta failed, error:" + api_response.text)
            result = api_response.text
            result_dict = json.loads(result)
            if result_dict['code'] != 0:
                raise requests.exceptions.RequestException(
                    "user:" + login_params.username + " login to clickzetta failed, error:" + result_dict['message'])
            if result_dict['data']['token'] is None:
                raise requests.exceptions.RequestException(
                    "user:" + login_params.username + " login to clickzetta  failed, error: token is None")
            else:
                token = result_dict['data']['token']
                return token
        except requests.exceptions.RequestException:
            raise

    def create_bulkload_stream(self, schema_name: string, table_name: string,
                               options: BulkLoadOptions) -> BulkLoadMetaData:
        create_bulk_load_request = ingestion_pb2.CreateBulkLoadStreamRequest()
        account = ingestion_pb2.Account()
        user_ident = ingestion_pb2.UserIdentifier()
        user_ident.instance_id = self.instance_id
        user_ident.workspace = self.workspace
        user_ident.user_name = self.username
        account.user_ident.CopyFrom(user_ident)
        account.token = self.token
        create_bulk_load_request.account.CopyFrom(account)
        table_identifier = ingestion_pb2.TableIdentifier()
        table_identifier.instance_id = self.instance_id
        table_identifier.workspace = self.workspace
        table_identifier.schema_name = schema_name
        table_identifier.table_name = table_name
        create_bulk_load_request.identifier.CopyFrom(table_identifier)
        if options.operation == BulkLoadOperation.APPEND:
            create_bulk_load_request.operation = ingestion_pb2.BulkLoadStreamOperation.BL_APPEND
        elif options.operation == BulkLoadOperation.UPSERT:
            create_bulk_load_request.operation = ingestion_pb2.BulkLoadStreamOperation.BL_UPSERT
        elif options.operation == BulkLoadOperation.OVERWRITE:
            create_bulk_load_request.operation = ingestion_pb2.BulkLoadStreamOperation.BL_OVERWRITE
        if options.partition_specs is not None:
            create_bulk_load_request.partition_spec = options.partition_specs
        if options.record_keys is not None:
            keys = []
            for key in options.record_keys:
                keys.append(key)
            create_bulk_load_request.record_keys.extend(keys)
        create_bulk_load_request.prefer_internal_endpoint = options.prefer_internal_endpoint
        response = self._gate_way_call(create_bulk_load_request, ingestion_pb2.MethodEnum.CREATE_BULK_LOAD_STREAM_V2)
        response_pb = ParseDict(response, ingestion_pb2.CreateBulkLoadStreamResponse(), ignore_unknown_fields=True)
        self.instance_id = response_pb.info.identifier.instance_id
        bulkload_meta_data = BulkLoadMetaData(response_pb.info.identifier.instance_id, response_pb.info)

        return bulkload_meta_data

    def commit_bulkload_stream(self, instance_id: int, workspace: string, schema_name: string, table_name: string,
                               stream_id: string, execute_workspace: string, execute_vc: string,
                               commit_mode: BulkLoadCommitMode) -> BulkLoadMetaData:
        commit_bulkload_request = ingestion_pb2.CommitBulkLoadStreamRequest()
        account = ingestion_pb2.Account()
        user_ident = ingestion_pb2.UserIdentifier()
        user_ident.instance_id = instance_id
        user_ident.workspace = workspace
        user_ident.user_name = self.username
        account.user_ident.CopyFrom(user_ident)
        account.token = self.token
        commit_bulkload_request.account.CopyFrom(account)
        table_identifier = ingestion_pb2.TableIdentifier()
        table_identifier.instance_id = instance_id
        table_identifier.workspace = workspace
        table_identifier.schema_name = schema_name
        table_identifier.table_name = table_name
        commit_bulkload_request.identifier.CopyFrom(table_identifier)
        commit_bulkload_request.stream_id = stream_id
        commit_bulkload_request.execute_workspace = execute_workspace
        commit_bulkload_request.execute_vc_name = execute_vc
        if commit_mode == BulkLoadCommitMode.COMMIT_STREAM:
            commit_bulkload_request.commit_mode = ingestion_pb2.CommitBulkLoadStreamRequest.CommitMode.COMMIT_STREAM
        elif commit_mode == BulkLoadCommitMode.ABORT_STREAM:
            commit_bulkload_request.commit_mode = ingestion_pb2.CommitBulkLoadStreamRequest.CommitMode.ABORT_STREAM

        response = self._gate_way_call(commit_bulkload_request, ingestion_pb2.MethodEnum.COMMIT_BULK_LOAD_STREAM_V2)
        response_pb = ParseDict(response, ingestion_pb2.CommitBulkLoadStreamResponse(), ignore_unknown_fields=True)
        bulkload_meta_data = BulkLoadMetaData(self.instance_id, response_pb.info)
        return bulkload_meta_data

    def get_bulkload_stream(self, schema_name: string, table_name: string, stream_id: string) -> BulkLoadMetaData:
        get_bulkload_stream_request = ingestion_pb2.GetBulkLoadStreamRequest()
        account = ingestion_pb2.Account()
        user_ident = ingestion_pb2.UserIdentifier()
        user_ident.instance_id = self.instance_id
        user_ident.workspace = self.workspace
        user_ident.user_name = self.username
        account.user_ident.CopyFrom(user_ident)
        account.token = self.token
        get_bulkload_stream_request.account.CopyFrom(account)
        table_identifier = ingestion_pb2.TableIdentifier()
        table_identifier.instance_id = self.instance_id
        table_identifier.workspace = self.workspace
        table_identifier.schema_name = schema_name
        table_identifier.table_name = table_name
        get_bulkload_stream_request.identifier.CopyFrom(table_identifier)
        get_bulkload_stream_request.stream_id = stream_id
        get_bulkload_stream_request.need_table_meta = True
        response = self._gate_way_call(get_bulkload_stream_request, ingestion_pb2.MethodEnum.GET_BULK_LOAD_STREAM_V2)
        response_pb = ParseDict(response, ingestion_pb2.GetBulkLoadStreamResponse(), ignore_unknown_fields=True)
        bulkload_meta_data = BulkLoadMetaData(self.instance_id, response_pb.info)
        return bulkload_meta_data

    def open_bulkload_stream_writer(self, instance_id: int, workspace: string, schema_name: string, table_name: string,
                                    stream_id: string, partition_id: int) -> BulkLoadConfig:
        open_bulkload_stream_request = ingestion_pb2.OpenBulkLoadStreamWriterRequest()
        account = ingestion_pb2.Account()
        user_ident = ingestion_pb2.UserIdentifier()
        user_ident.instance_id = instance_id
        user_ident.workspace = workspace
        user_ident.user_name = self.username
        account.user_ident.CopyFrom(user_ident)
        account.token = self.token
        open_bulkload_stream_request.account.CopyFrom(account)
        table_identifier = ingestion_pb2.TableIdentifier()
        table_identifier.instance_id = instance_id
        table_identifier.workspace = workspace
        table_identifier.schema_name = schema_name
        table_identifier.table_name = table_name
        open_bulkload_stream_request.identifier.CopyFrom(table_identifier)
        open_bulkload_stream_request.stream_id = stream_id
        open_bulkload_stream_request.partition_id = partition_id
        response = self._gate_way_call(open_bulkload_stream_request,
                                       ingestion_pb2.MethodEnum.OPEN_BULK_LOAD_STREAM_WRITER_V2)
        response_pb = ParseDict(response, ingestion_pb2.OpenBulkLoadStreamWriterResponse(), ignore_unknown_fields=True)
        bulkload_config = response_pb.config
        return BulkLoadConfig(bulkload_config)

    def finish_bulkload_stream_writer(self, instance_id: int, workspace: string, schema_name: string,
                                      table_name: string,
                                      stream_id: string, partition_id: int, written_files: list,
                                      written_lengths: list) -> ingestion_pb2.ResponseStatus:
        finish_bulkload_stream_request = ingestion_pb2.FinishBulkLoadStreamWriterRequest()
        account = ingestion_pb2.Account()
        user_ident = ingestion_pb2.UserIdentifier()
        user_ident.instance_id = instance_id
        user_ident.workspace = workspace
        user_ident.user_name = self.username
        account.user_ident.CopyFrom(user_ident)
        account.token = self.token
        finish_bulkload_stream_request.account.CopyFrom(account)
        table_identifier = ingestion_pb2.TableIdentifier()
        table_identifier.instance_id = instance_id
        table_identifier.workspace = workspace
        table_identifier.schema_name = schema_name
        table_identifier.table_name = table_name
        finish_bulkload_stream_request.identifier.CopyFrom(table_identifier)
        finish_bulkload_stream_request.stream_id = stream_id
        finish_bulkload_stream_request.partition_id = partition_id
        finish_bulkload_stream_request.written_files.extend(written_files)
        finish_bulkload_stream_request.written_lengths.extend(written_lengths)
        response = self._gate_way_call(finish_bulkload_stream_request,
                                       ingestion_pb2.MethodEnum.FINISH_BULK_LOAD_STREAM_WRITER_V2)
        response_pb = ParseDict(response, ingestion_pb2.FinishBulkLoadStreamWriterResponse(), ignore_unknown_fields=True)
        return response_pb.status

    def _gate_way_call(self, request, method: ingestion_pb2.MethodEnum):
        path = '/igs/gatewayEndpoint'
        gate_way_request = ingestion_pb2.GatewayRequest()
        gate_way_request.methodEnumValue = method
        gate_way_request.message = MessageToJson(request)

        HEADERS['instanceName'] = self.instance
        HEADERS['X-ClickZetta-Token'] = self.token
        try:
            api_response = requests.post(self.service + path, data=MessageToJson(gate_way_request), headers=HEADERS)
            api_response.encoding = 'utf-8'
            result = api_response.text
            result_dict = json.loads(result)
            if api_response.status_code != 200:
                raise requests.exceptions.RequestException(
                    'gate_way_call return failed code.Error message:' + result_dict['message'])
            result_status = ParseDict(result_dict['status'], ingestion_pb2.GateWayResponseStatus)
            if result_status.code == ingestion_pb2.Code.SUCCESS:
                message_json = json.loads(result_dict['message'])
                internal_result_status = message_json['status']
                if internal_result_status["code"] == "SUCCESS":
                    return json.loads(result_dict['message'])
                else:
                    raise requests.exceptions.RequestException(
                        'gate_way_call return failed code.Error message:' + internal_result_status["error_message"])
            else:
                raise requests.exceptions.RequestException(
                    'gate_way_call return failed code.Error message:' + result_status.message)

        except requests.exceptions.RequestException as request_exception:
            logging.error('gate_way_request error:{}'.format(request_exception))
            raise requests.exceptions.RequestException(
                'gate_way_request error:{}'.format(request_exception))
        except Exception as e:
            logging.error('gate_way_request error:{}'.format(e))
            raise requests.exceptions.RequestException(
                'gate_way_request error:{}'.format(e))

    def pre_process_sql(self, sql):
        queries = split_sql(sql)
        if len(queries) > 0:
            for query in queries[:-1]:
                self.process_use_cmd(query)
        return queries[-1] + '\n;'

    def process_use_cmd(self, sql: str):
        sql = re.sub(r'\s+', ' ', sql)
        if sql.lower().startswith("use vcluster "):
            self.vcluster = sql.lower().replace('use vcluster ', '').strip()
        elif sql.lower().startswith('use schema '):
            self.schema = sql.lower().replace('use schema ', '').strip()
        elif sql.lower().startswith("use "):
            self.schema = sql.lower().replace('use ', '').strip()

    def submit_sql_job(self, token: str, sql: str, job_id: JobID, parameters=None, schema=None) -> QueryResult:
        sql = self.pre_process_sql(sql)
        path = "/lh/submitJob"
        if 'test plain returns' in sql or 'test unicode returns' in sql:
            return QueryResult({})
        table = Table(self.workspace, '', self.instance, self.vcluster)
        vc = table.vcluster
        job_type = JobType.SQL_JOB
        logging.info(
            "clickzetta connector submitting job,  id:" + job_id.id)
        job_name = "SQL_JOB"
        user_id = 0
        reqeust_mode = JobRequestMode.HYBRID
        timeout = FIRST_SQL_SUBMIT_TIMEOUT
        job_config = {}
        sdk_job_timeout = 0
        sdk_job_priority = 15
        sql_job_set_prop = {'cz.sql.adhoc.result.type': 'embedded', 'cz.sql.adhoc.default.format': 'arrow'}
        job_timeout_ms = 0
        hints = self.hints if self.hints is not None else {}
        if parameters is not None and len(parameters) != 0:
            if 'hints' in parameters.keys():
                hints = parameters['hints']
            for key in parameters.keys():
                if "%(" + key + ")s" in sql:
                    sql = sql.replace("%(" + key + ")s", str(parameters[key]))
        if len(hints) > 0:
            for key in hints:
                if key == 'sdk.job.timeout':
                    sdk_job_timeout = int(hints[key])
                elif key == 'sdk.job.priority':
                    sdk_job_priority = int(hints[key])
                elif key == 'sdk.query.timeout.ms':
                    job_timeout_ms = int(hints[key])
                else:
                    sql_job_set_prop[key] = hints[key]
        timeout = sdk_job_timeout if 0 < sdk_job_timeout <= timeout else timeout
        sql_config = SQLJobConfig(0, "0", "0", sql_job_set_prop)
        schema_name = ''
        if self.schema is not None:
            schema_name = self.schema
        if schema is not None:
            schema_name = schema
        sql_job = SQLJob(sql_config, table.workspace, schema_name, [sql])
        user_agent = ""

        job_desc = JobDesc(vc, job_type, job_id, job_name, user_id, reqeust_mode, timeout, job_config, sql_job,
                           job_timeout_ms,
                           user_agent, sdk_job_priority)
        job_request = JobRequest(job_desc)
        data = json.dumps(job_request.to_api_repr())
        HEADERS['instanceName'] = table.instance
        HEADERS['X-ClickZetta-Token'] = token
        logging.debug('BEGIN TO SEND REQUEST:' + sql + ' TO CZ SERVER, TIME:' + str(datetime.now()))
        try:
            api_response = self.session.post(self.service + path, data=data, headers=HEADERS)
            if api_response.status_code != 200:
                raise requests.exceptions.RequestException(
                    "submit sql job failed:" + str(api_response.status_code) + api_response.text + ".sql:" + sql)
            api_response.encoding = 'utf-8'
            result = api_response.text
            result_dict = json.loads(result)
            if int(result_dict['code']) != 200:
                raise requests.exceptions.RequestException(
                    "submit sql job failed:" + str(result_dict))
            get_job_result_dict = self.process_sql_response_with_timeout(sdk_job_timeout, result_dict, job_id,
                                                                         job_timeout_ms)
            logging.debug('GET RESPONSE FROM CZ SERVER FOR REQUEST:' + sql + ' TIME:' + str(datetime.now()))
            query_result = QueryResult(get_job_result_dict)
            if sql.lstrip().upper().startswith('PUT ') or sql.lstrip().upper().startswith('GET '):
                check_volume_result = self.process_volume_sql(token, sql, query_result)
                return check_volume_result
            return query_result
        except Exception as e:
            raise Exception("submit sql job failed:" + str(e))

    def process_volume_sql(self, token: str, sql: str, query_result: QueryResult) -> QueryResult:
        upper_sql = sql.strip().upper()
        outcome = query_result.data.read()
        if len(outcome) != 1:
            raise Exception(f"get volume sql failed, with result: {outcome}")
        outcome_obj = json.loads(outcome[0][0])
        if outcome_obj['status'] == 'FAILED':
            raise Exception(f"{outcome_obj['request']['command']} volume failed: {outcome_obj['error']}")
        if upper_sql.startswith('GET'):
            query_result = self.gen_volume_result(outcome_obj)
            if outcome_obj['status'] == 'CONTINUE' and outcome_obj['nextMarker'] != '':
                sql = 'set cz.sql.volume.file.transfer.next.marker=nextMarker;' + sql
                job_id = JobID(self._format_job_id(), self.workspace, DEFAULT_INSTANCE_ID)
                next_result = self.submit_sql_job(token=token, sql=sql, job_id=job_id)
                return self._merge_query_result(query_result, next_result)
            return query_result
        elif upper_sql.startswith("PUT"):
            if outcome_obj['status'] == 'CONTINUE':
                sql = self._gen_new_put_sql(outcome_obj)
                job_id = JobID(self._format_job_id(), self.workspace, DEFAULT_INSTANCE_ID)
                return self.submit_sql_job(token=token, sql=sql, job_id=job_id)
            else:
                return self.gen_volume_result(outcome_obj)

    def gen_volume_result(self, outcome: Any):
        result_list = []
        if outcome['request']['command'].upper() == 'GET':
            volume_files = self.get_volume_files(outcome)
            for i in range(len(volume_files)):
                response = requests.get(outcome['ticket']['presignedUrls'][i], stream=True)
                local_path = os.path.join(outcome['request']['localPaths'][0], volume_files[i])
                if not os.path.exists(os.path.dirname(local_path)):
                    os.makedirs(os.path.dirname(local_path))
                with open(local_path, 'wb') as out_file:
                    shutil.copyfileobj(response.raw, out_file)
                logging.info(f"get volume success, volume_path:{volume_files[i]}, local_path:{local_path}")
                res = [volume_files[i], local_path, os.path.getsize(local_path)]
                result_list.append(res)
        elif outcome['request']['command'].upper() == 'PUT':
            if outcome['status'] == 'SUCCESS':
                for i in range(len(outcome['ticket']['presignedUrls'])):
                    local_path = outcome['request']['localPaths'][i]
                    if not os.path.exists(local_path):
                        raise Exception(f"put volume failed, local_path:{local_path} not exists")
                    with open(local_path, 'rb') as f:
                        requests.put(outcome['ticket']['presignedUrls'][i], data=f.read(), headers={})
                    logging.info(f"put volume success, local_path:{local_path}")
                    volume_file = ''
                    if 'file' in outcome['request']:
                        volume_file = outcome['request']['file']
                    else:
                        if 'subdirectory' in outcome['request']:
                            path = outcome['request']['subdirectory']
                            path = path if path[-1] == '/' else path + '/'
                            volume_file = path + os.path.basename(local_path)
                        else:
                            volume_file = os.path.basename(local_path)
                    res = [local_path, volume_file, os.path.getsize(local_path)]
                    result_list.append(res)
        query_result = QueryResult({})
        query_result.data = QueryData(data=result_list, data_type=QueryDataType.Memory)
        return query_result

    def get_volume_files(self, outcome: Any):
        volume_files = []
        if 'file' in outcome['request']:
            volume_files.append(os.path.basename(os.path.abspath(outcome['request']['file'])))
        else:
            for url in outcome['ticket']['presignedUrls']:
                parsed_url = urlparse(unquote(url))
                path_parts = parsed_url.path.split('/')
                volume_files.append(path_parts[-1])
        return volume_files

    def process_sql_response_with_timeout(self, sdk_job_timeout, result_dict, job_id, job_timeout_ms):
        if ('status' in result_dict and 'errorCode' in result_dict['status'] and
                result_dict['status']['errorCode'] == 'CZLH-60010'):
            raise Exception("submit sql job:" + str(job_id.id) + " timeout after " + str(job_timeout_ms)
                            + " ms. killed by lakehouse")
        if 0 < sdk_job_timeout <= FIRST_SQL_SUBMIT_TIMEOUT:
            # check resultSet first, then check status
            if 'resultSet' in result_dict:
                if 'data' in result_dict['resultSet']:
                    if 'data' in result_dict['resultSet']['data']:
                        return result_dict
            if 'status' in result_dict:
                if 'state' in result_dict['status']:
                    status = result_dict['status']['state']
                    if status == "QUEUEING" or status == "SETUP" or status == "RUNNING":
                        raise Exception("submit sql job:" + str(job_id.id) + " timeout after " + str(
                            FIRST_SQL_SUBMIT_TIMEOUT) + " seconds. killed by sdk.")
                    else:
                        return result_dict
            else:
                raise Exception("submit sql job:" + str(job_id.id) + " failed, root: " + str(result_dict))

        elif sdk_job_timeout > FIRST_SQL_SUBMIT_TIMEOUT:
            return self.wait_job_finished(result_dict, job_id, HEADERS, sdk_job_timeout - FIRST_SQL_SUBMIT_TIMEOUT)
        else:
            return self.wait_job_finished(result_dict, job_id, HEADERS, HYBRID_SQL_TIMEOUT - FIRST_SQL_SUBMIT_TIMEOUT)

    def close(self):
        pass

    def cancel_job(self, job_id: JobID, headers):
        account = clickzetta.enums.ClickZettaAccount(0)
        cancel_job_request = clickzetta.enums.CancelJobRequest(account, job_id, '', False)
        path = '/lh/cancelJob'
        data = json.dumps(cancel_job_request.to_api_repr())
        try:
            self.session.post(self.service + path, data=data, headers=headers)
        except Exception as e:
            logging.error('clickzeta connector cancel job error:{}'.format(e))

    def check_if_job_timeout(self, sdk_job_timeout, sql_job_start_time, job_id, headers):
        if sdk_job_timeout > 0:
            if (datetime.now() - sql_job_start_time).seconds > sdk_job_timeout:
                self.cancel_job(job_id, headers)
                return True
        return False

    def get_job_profile(self, job_id: string):
        result_dict = self._get_job(job_id, clickzetta.enums.JobRequestType.PROFILE)
        return result_dict

    def get_job_result(self, job_id: string):
        result_dict = self._get_job(job_id, clickzetta.enums.JobRequestType.RESULT)
        return result_dict

    def get_job_progress(self, job_id: string):
        result_dict = self._get_job(job_id, clickzetta.enums.JobRequestType.PROGRESS)
        return result_dict

    def get_job_summary(self, job_id: string):
        result_dict = self._get_job(job_id, clickzetta.enums.JobRequestType.SUMMARY)
        return result_dict

    def get_job_plan(self, job_id: string):
        result_dict = self._get_job(job_id, clickzetta.enums.JobRequestType.PLAN)
        return result_dict

    def _get_job(self, job_id: string, type: string):
        HEADERS['instanceName'] = self.instance
        HEADERS['X-ClickZetta-Token'] = self.token
        id = JobID(job_id, self.workspace, DEFAULT_INSTANCE_ID)
        account = clickzetta.enums.ClickZettaAccount(0)
        get_job_request = clickzetta.enums.GetJobRequest(account, id, 0, '')
        path = '/lh/getJob'
        try:
            api_request = clickzetta.enums.APIGetJobRequest(get_job_request, '', type)
            data = json.dumps(api_request.to_api_repr())
            api_response = self.session.post(self.service + path, data=data, headers=HEADERS)
            result = api_response.text
            result_dict = json.loads(result)
            if int(result_dict['code']) != 200:
                raise requests.exceptions.RequestException(f"Get job:{id.id} {type} failed:{result_dict}")
            return result_dict
        except AttributeError as e:
            raise requests.exceptions.RequestException(f"Get job:{id.id} {type} failed:{e}, Only supports：result/progress/summary/plan/profile.")
        except Exception as e:
            raise Exception(f"Get job:{id.id} {type} failed:{e}")

    def is_job_finished(self, result_dict):
        if 'status' in result_dict:
            if 'state' in result_dict['status']:
                status = result_dict['status']['state']
                if status == "SUCCEED" or status == 'FAILED':
                    return True
                elif status == "QUEUEING" or status == "SETUP" or status == "RUNNING":
                    if 'resultSet' in result_dict:
                        if 'data' in result_dict['resultSet']:
                            if 'data' in result_dict['resultSet']['data']:
                                return True
                        elif 'location' in result_dict['resultSet']:
                            return True
        return False

    def wait_job_finished(self, result_dict, job_id: JobID, headers, sdk_job_timeout=0):
        if 'status' in result_dict:
            if 'state' in result_dict['status']:
                if not self.is_job_finished(result_dict):
                    account = clickzetta.enums.ClickZettaAccount(0)
                    get_job_result_request = clickzetta.enums.GetJobRequest(account, job_id, 0, '')
                    api_request = clickzetta.enums.APIGetJobRequest(get_job_result_request, '', clickzetta.enums.JobRequestType.RESULT)
                    path = '/lh/getJob'
                    data = json.dumps(api_request.to_api_repr())
                    sql_job_start_time = datetime.now()
                    while True:
                        if self.check_if_job_timeout(sdk_job_timeout, sql_job_start_time, job_id, headers):
                            logging.error("clickzetta sql job:" + str(job_id.id) + " timeout after " + str(
                                sdk_job_timeout) + " seconds. killed by sdk.")
                            raise Exception("clickzetta sql job:" + str(job_id.id) + " timeout after " + str(
                                sdk_job_timeout) + " seconds. killed by sdk.")
                        get_job_result_response = self.session.post(self.service + path, data=data, headers=headers)
                        result = get_job_result_response.text
                        get_job_result_dict = json.loads(result)
                        if int(get_job_result_dict['code']) != 200:
                            raise requests.exceptions.RequestException(
                                "Get async sql job:" + str(job_id.id) + " result failed:" + get_job_result_dict)
                        if 'status' in get_job_result_dict:
                            if 'state' in get_job_result_dict['status']:
                                if get_job_result_dict['status']['state'] == 'SUCCEED' or \
                                        get_job_result_dict['status']['state'] == 'FAILED' or \
                                        get_job_result_dict['status']['state'] == 'CANCELLED':
                                    logging.info(
                                        "Get async sql job:" + str(job_id.id) + " result successfully.")
                                    return get_job_result_dict
                                else:
                                    logging.info("Get async sql job:" + str(
                                        job_id.id) + " result pending, retry after 10 seconds.")
                                    time.sleep(3)
                            else:
                                raise Exception(
                                    "Get async sql job:" + str(
                                        job_id.id) + " result failed, http response job state is not exist.")
                        else:
                            raise Exception(
                                "Get async sql job:" + str(
                                    job_id.id) + " result failed, http response job status is not exist.")

                else:
                    return result_dict
            else:
                raise Exception("Execute sql job:" + str(job_id.id) + " failed, http response job state is not exist.")
        else:
            raise Exception("Execute sql job:" + str(job_id.id) + " failed, http response job status is not exist.")

    def _format_job_id(self):
        unique_id = str(datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f'))
        format_unique_id = unique_id.replace('-', '').replace(':', '').replace('.', '').replace(' ', '') \
                           + str(random.randint(10000, 99999))
        return format_unique_id

    def generate_job_id(self):
        return self._format_job_id()

    def get_table_names(self, schema: str):
        job_id = JobID(self._format_job_id(), self.workspace, DEFAULT_INSTANCE_ID)
        query_result = self.submit_sql_job(token=self.token, sql='show tables;', schema=schema, job_id=job_id)
        query_data = query_result.data.read()
        table_names = []
        for entry in query_data:
            table_names.append(entry[1])

        return table_names

    def get_schemas(self):
        job_id = JobID(self._format_job_id(), self.workspace, DEFAULT_INSTANCE_ID)
        query_result = self.submit_sql_job(token=self.token, sql='show schemas;', job_id=job_id)
        query_data = query_result.data.read()
        schema_names = []
        for entry in query_data:
            schema_names.append(entry[0])

        return schema_names

    def get_columns(self, table_name: str, schema: str):
        job_id = JobID(self._format_job_id(), self.workspace, DEFAULT_INSTANCE_ID)
        if '.' in table_name:
            schema = table_name.split('.')[0]
            table_name = table_name.split('.')[1]
        query_result = self.submit_sql_job(token=self.token,
                                           sql='select * from ' + schema + '.' + table_name + ' limit 1;',
                                           job_id=job_id)
        schema = query_result.schema

        return schema

    def has_table(self, full_table_name: str):
        job_id = JobID(self._format_job_id(), self.workspace, DEFAULT_INSTANCE_ID)
        try:
            query_result = self.submit_sql_job(token=self.token, sql='show create table ' + full_table_name + ';',
                                               job_id=job_id)
        except Exception as e:
            return False
        if query_result.state != 'FAILED':
            return True
        else:
            return False

    def _merge_query_result(self, res1: QueryResult, res2: QueryResult):
        new_data = res1.data.data + res2.data.data
        res1.data = QueryData(data=new_data, data_type=QueryDataType.Memory)
        return res1

    def _gen_new_put_sql(self, outcome_obj):
        prefix = ''
        option_part = ''
        for option in outcome_obj['request']['options']:
            if option['name'].upper() == 'SOURCE_PREFIX':
                prefix = option['value']
            option_part += ' ' + option['name'] + ' = ' + option['value']
        requested_paths = outcome_obj['request']['localPaths']
        resolved_paths = []
        for p in requested_paths:
            resolved_paths += resolve_local_path(p)
        if len(resolved_paths) == 0:
            raise Exception('No local file to put into volume')
        sql = 'PUT '
        sql += ", ".join(f"'{p}'" for p in resolved_paths)
        sql += ' TO ' + outcome_obj['request']['volumeIdentifier']
        if 'subdirectory' in outcome_obj['request']:
            sql += " SUBDIRECTORY '" + outcome_obj['request']['subdirectory'] + "'"
        elif 'file' in outcome_obj['request']:
            sql += " FILE '" + outcome_obj['request']['file'] + "'"
        sql += option_part
        sql += ';'
        return sql


def _add_server_timeout_header(headers: Optional[Dict[str, str]], kwargs):
    timeout = kwargs.get("timeout")
    if timeout is not None:
        if headers is None:
            headers = {}
        headers[TIMEOUT_HEADER] = str(timeout)

    if headers:
        kwargs["headers"] = headers

    return kwargs



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: property.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0eproperty.proto\x12\x08\x63z.proto\"&\n\x08Property\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"4\n\nProperties\x12&\n\nproperties\x18\x01 \x03(\x0b\x32\x12.cz.proto.Propertyb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'property_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_PROPERTY']._serialized_start=28
  _globals['_PROPERTY']._serialized_end=66
  _globals['_PROPERTIES']._serialized_start=68
  _globals['_PROPERTIES']._serialized_end=120
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: share_meta.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import object_identifier_pb2 as object__identifier__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10share_meta.proto\x12\x08\x63z.proto\x1a\x17object_identifier.proto\"\xce\x01\n\x05Share\x12\x36\n\x12provider_workspace\x18\x01 \x01(\x0b\x32\x1a.cz.proto.ObjectIdentifier\x12\"\n\x04kind\x18\x02 \x01(\x0e\x32\x14.cz.proto.Share.Kind\x12$\n\x05scope\x18\x03 \x01(\x0e\x32\x15.cz.proto.Share.Scope\"!\n\x04Kind\x12\x0b\n\x07INBOUND\x10\x00\x12\x0c\n\x08OUTBOUND\x10\x01\" \n\x05Scope\x12\x0b\n\x07PRIVATE\x10\x00\x12\n\n\x06PUBLIC\x10\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'share_meta_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SHARE']._serialized_start=56
  _globals['_SHARE']._serialized_end=262
  _globals['_SHARE_KIND']._serialized_start=195
  _globals['_SHARE_KIND']._serialized_end=228
  _globals['_SHARE_SCOPE']._serialized_start=230
  _globals['_SHARE_SCOPE']._serialized_end=262
# @@protoc_insertion_point(module_scope)



import copy
import enum
import string

DEFAULT_NS = ['default', 'lh']


class AutoRowIDs(enum.Enum):
    DISABLED = enum.auto()
    GENERATE_UUID = enum.auto()
class FetchMode(enum.Enum):
    FETCH_ONE = enum.auto()
    FETCH_ALL = enum.auto()
    FETCH_MANY = enum.auto()

class Compression(object):
    GZIP = "GZIP"

    DEFLATE = "DEFLATE"

    SNAPPY = "SNAPPY"

    NONE = "NONE"


class DecimalTargetType:
    STRING = "STRING"


class CreateDisposition(object):
    CREATE_IF_NEEDED = "CREATE_IF_NEEDED"

    CREATE_NEVER = "CREATE_NEVER"


class DestinationFormat(object):
    CSV = "CSV"

    NEWLINE_DELIMITED_JSON = "NEWLINE_DELIMITED_JSON"

    AVRO = "AVRO"

    PARQUET = "PARQUET"


class Encoding(object):
    UTF_8 = "UTF-8"

    ISO_8859_1 = "ISO-8859-1"


class QueryPriority(object):
    INTERACTIVE = "INTERACTIVE"

    BATCH = "BATCH"


class QueryApiMethod(str, enum.Enum):
    SELECT = "SELECT"

    SHOW = "SHOW"

    DROP = "DROP"

    ALTER = "ALTER"

    CREATE = "CREATE"

    TRUNCATE = "TRUNCATE"


class JobType(object):
    SQL_JOB = "SQL_JOB"

    COMPACTION_JOB = "COMPACTION_JOB"


class JobStatus(object):
    SETUP = "SETUP"
    QUEUEING = "QUEUEING"
    RUNNING = "RUNNING"
    SUCCEED = "SUCCEED"
    CANCELLING = "CANCELLING"
    CANCELLED = "CANCELLED"
    FAILED = "FAILED"


class JobRequestMode(object):
    UNKNOWN = "UNKNOWN"
    HYBRID = "HYBRID"
    ASYNC = "ASYNC"
    SYNC = "SYNC"


class JobRequestType(object):
    PLAN = "get_plan_request"
    PROFILE = "get_profile_request"
    RESULT = "get_result_request"
    PROGRESS = "get_progress_request"
    SUMMARY = "get_summary_request"


class SQLJobConfig(object):
    def __init__(self, timeout, adhoc_size_limit: string, adhoc_row_limit: string, hint: map):
        self.timeout = timeout
        self.adhoc_size_limit = adhoc_size_limit
        self.adhoc_row_limit = adhoc_row_limit
        self.hint = hint
        self._properties = {'timeout': timeout, 'adhocSizeLimit': adhoc_size_limit, 'adhocRowLimit': adhoc_row_limit,
                            'hint': hint}

    def to_api_repr(self) -> dict:
        return copy.deepcopy(self._properties)


class SQLJob(object):
    def __init__(self, sql_job_config: SQLJobConfig, workspace: string, db_name: string, query):
        self.query = query
        self.default_namespace = [workspace, db_name]
        self.sql_job_config = sql_job_config

    def to_api_repr(self) -> dict:
        _properties = {'query': self.query, 'defaultNamespace': self.default_namespace,
                       'sqlConfig': self.sql_job_config.to_api_repr()}
        return copy.deepcopy(_properties)


class JobID(object):
    def __init__(self, id: string, workspace: string, instance_id: int):
        self.id = id
        self.workspace = workspace
        self.instance_id = instance_id
        self._properties = {'id': id, 'workspace': workspace, 'instance_id': instance_id}

    def to_api_repr(self) -> dict:
        return copy.deepcopy(self._properties)


class ClickZettaAccount(object):
    def __init__(self, user_id: int):
        self.user_id = user_id
        self._properties = {'user_id': user_id}

    def to_api_repr(self) -> dict:
        return copy.deepcopy(self._properties)


class GetJobRequest(object):
    def __init__(self, account: ClickZettaAccount, job_id: JobID, offset: int, user_agent: string):
        self.account = account
        self.job_id = job_id
        self.offset = offset
        self.user_agent = user_agent
        self._properties = {'account': account.to_api_repr(), 'job_id': job_id.to_api_repr(), 'offset': offset,
                            'user_agent': user_agent}

    def to_api_repr(self) -> dict:
        return copy.deepcopy(self._properties)


class CancelJobRequest(object):
    def __init__(self, account: ClickZettaAccount, job_id: JobID, user_agent: string, force: bool):
        self.account = account
        self.job_id = job_id
        self.user_agent = user_agent
        self.force = force
        self._properties = {'account': account.to_api_repr(), 'job_id': job_id.to_api_repr(), 'user_agent': user_agent,
                            'force': force}

    def to_api_repr(self) -> dict:
        return copy.deepcopy(self._properties)


class GetJobProfileRequest(object):
    def __init__(self, account: ClickZettaAccount, job_id: JobID, user_agent: string):
        self.account = account
        self.job_id = job_id
        self.user_agent = user_agent
        self._properties = {'account': account.to_api_repr(), 'job_id': job_id.to_api_repr(), 'user_agent': user_agent}

    def to_api_repr(self) -> dict:
        return copy.deepcopy(self._properties)


class APIGetJobProfileRequest(object):
    def __init__(self, get_job_profile_request: GetJobProfileRequest, user_agent: string):
        self.get_job_profile_request = get_job_profile_request
        self.user_agent = user_agent
        self._properties = {'get_profile_request': get_job_profile_request.to_api_repr(), 'user_agent': user_agent}

    def to_api_repr(self) -> dict:
        return copy.deepcopy(self._properties)


class APIGetJobRequest(object):
    def __init__(self, get_job_request: GetJobRequest, user_agent: string, job_request_type: string):
        self.get_job_result_request = get_job_request
        self.user_agent = user_agent
        self._properties = {f'{job_request_type}': get_job_request.to_api_repr(), 'user_agent': user_agent}

    def to_api_repr(self) -> dict:
        return copy.deepcopy(self._properties)


class JobDesc(object):
    def __init__(self, virtual_cluster: string, job_type: JobType, job_id: JobID, job_name: string, user_id: int,
                 request_mode: JobRequestMode, hybrid_polling_timeout: int, job_config: map, sql_job: SQLJob,
                 job_timeout_ms: int, user_agent: string, priority: int):
        self.virtual_cluster = virtual_cluster
        self.job_type = job_type
        self.job_id = job_id
        self.job_name = job_name
        self.user_id = user_id
        self.request_mode = request_mode
        self.hybrid_polling_timeout = hybrid_polling_timeout
        self.job_config = job_config
        self.sql_job = sql_job
        self.job_timeout_ms = job_timeout_ms
        self.user_agent = user_agent
        self.priority = priority
        self._properties = {'virtualCluster': virtual_cluster, 'type': job_type, 'jobId': job_id.to_api_repr(),
                            'jobName': job_name,
                            'requestMode': request_mode, 'hybridPollingTimeout': hybrid_polling_timeout,
                            'jobConfig': job_config, 'sqlJob': sql_job.to_api_repr()}
        if 'query_tag' in sql_job.sql_job_config.hint:
            self._properties['query_tag'] = sql_job.sql_job_config.hint['query_tag']
        if job_timeout_ms > 0:
            self._properties['jobTimeoutMs'] = job_timeout_ms

    def to_api_repr(self) -> dict:
        return copy.deepcopy(self._properties)


class JobRequest(object):
    def __init__(self, job_desc: JobDesc):
        self.job_desc = job_desc
        self._properties = {"jobDesc": job_desc.to_api_repr()}

    def to_api_repr(self) -> dict:
        return copy.deepcopy(self._properties)


class LoginParams(object):
    def __init__(self, username: string, password: string, instance: string):
        self.username = username
        self.password = password
        self.instance = instance
        self._properties = {"username": username, "password": password, "instanceName": instance}

    def to_api_repr(self) -> dict:
        return copy.deepcopy(self._properties)


class ErrorStatus(object):
    Unauthorized = "UNAUTHORIZED"


class SourceFormat(object):
    CSV = "CSV"

    NEWLINE_DELIMITED_JSON = "NEWLINE_DELIMITED_JSON"

    AVRO = "AVRO"

    PARQUET = "PARQUET"

    ORC = "ORC"


class KeyResultStatementKind:
    KEY_RESULT_STATEMENT_KIND_UNSPECIFIED = "KEY_RESULT_STATEMENT_KIND_UNSPECIFIED"
    LAST = "LAST"
    FIRST_SELECT = "FIRST_SELECT"


class StandardSqlTypeNames(str, enum.Enum):
    def _generate_next_value_(name, start, count, last_values):
        return name

    TYPE_KIND_UNSPECIFIED = enum.auto()
    INT64 = enum.auto()
    BOOL = enum.auto()
    FLOAT64 = enum.auto()
    FLOAT32 = enum.auto()
    STRING = enum.auto()
    VARCHAR = enum.auto()
    CHAR = enum.auto()
    INT32 = enum.auto()
    INT16 = enum.auto()
    INT8 = enum.auto()
    TIMESTAMP = enum.auto()
    DATE = enum.auto()
    DECIMAL = enum.auto()
    BINARY = enum.auto()
    ARRAY = enum.auto()
    MAP = enum.auto()
    STRUCT = enum.auto()


class SqlTypeNames(str, enum.Enum):
    STRING = "STRING"
    BINARY = "BINARY"
    INT8 = "INT8"
    INT16 = "INT16"
    INT32 = "INT32"
    INT64 = "INT64"
    FLOAT32 = "FLOAT32"
    FLOAT64 = "FLOAT32"
    DECIMAL = "DECIMAL"
    BOOL = "BOOLEAN"
    STRUCT = "STRUCT"
    MAP = "MAP"
    ARRAY = "ARRAY"
    TIMESTAMP = "TIMESTAMP"
    DATE = "DATE"


class WriteDisposition(object):
    WRITE_APPEND = "WRITE_APPEND"

    WRITE_TRUNCATE = "WRITE_TRUNCATE"

    WRITE_EMPTY = "WRITE_EMPTY"



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: function_meta.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13\x66unction_meta.proto\x12\x08\x63z.proto\"-\n\x10\x46unctionResource\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0b\n\x03uri\x18\x02 \x01(\t\"E\n\x14\x46unctionResourceList\x12-\n\tresources\x18\x01 \x03(\x0b\x32\x1a.cz.proto.FunctionResource\"\xd2\x01\n\x10RemoteEntrypoint\x12\x14\n\x0cinternal_url\x18\x01 \x01(\t\x12\x14\n\x0c\x65xternal_url\x18\x02 \x01(\t\x12\x10\n\x08protocol\x18\x03 \x01(\t\x12\x13\n\x0bvendor_type\x18\x04 \x01(\t\x12:\n\x0bvendor_info\x18\x05 \x01(\x0b\x32%.cz.proto.RemoteEntrypoint.VendorInfo\x1a/\n\nVendorInfo\x12\x0f\n\x07service\x18\x01 \x01(\t\x12\x10\n\x08\x66unction\x18\x02 \x01(\t\"\xd4\x01\n\x08\x46unction\x12\x10\n\x08\x63\x61tegory\x18\x01 \x01(\t\x12\x11\n\texec_type\x18\x02 \x01(\t\x12\x11\n\tsignature\x18\x03 \x01(\t\x12\x0f\n\x07handler\x18\x04 \x01(\t\x12\x15\n\rconnection_id\x18\x05 \x01(\x03\x12\x31\n\tresources\x18\x06 \x01(\x0b\x32\x1e.cz.proto.FunctionResourceList\x12\x35\n\x11remote_entrypoint\x18\x07 \x01(\x0b\x32\x1a.cz.proto.RemoteEntrypointb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'function_meta_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_FUNCTIONRESOURCE']._serialized_start=33
  _globals['_FUNCTIONRESOURCE']._serialized_end=78
  _globals['_FUNCTIONRESOURCELIST']._serialized_start=80
  _globals['_FUNCTIONRESOURCELIST']._serialized_end=149
  _globals['_REMOTEENTRYPOINT']._serialized_start=152
  _globals['_REMOTEENTRYPOINT']._serialized_end=362
  _globals['_REMOTEENTRYPOINT_VENDORINFO']._serialized_start=315
  _globals['_REMOTEENTRYPOINT_VENDORINFO']._serialized_end=362
  _globals['_FUNCTION']._serialized_start=365
  _globals['_FUNCTION']._serialized_end=577
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pb_util.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rpb_util.proto\x12\x04kudu\x1a google/protobuf/descriptor.proto\"[\n\x14\x43ontainerSupHeaderPB\x12\x32\n\x06protos\x18\x01 \x02(\x0b\x32\".google.protobuf.FileDescriptorSet\x12\x0f\n\x07pb_type\x18\x02 \x02(\t:6\n\x06REDACT\x12\x1d.google.protobuf.FieldOptions\x18\xd1\x86\x03 \x01(\x08:\x05\x66\x61lseB\x11\n\x0forg.apache.kudu')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pb_util_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\017org.apache.kudu'
  _globals['_CONTAINERSUPHEADERPB']._serialized_start=57
  _globals['_CONTAINERSUPHEADERPB']._serialized_end=148
# @@protoc_insertion_point(module_scope)



from pathlib import Path
from typing import List
from os import walk


def _resolve_glob(path: str) -> List[Path]:
    # TODO(guantao.gao) support Windows path
    if path.startswith("/"):
        return [p.absolute() for p in Path("/").glob(path[1:])]
    else:
        return list(Path(".").glob(path))


def _resolve_dir(path: str) -> List[Path]:
    res = []
    for root, _, files in walk(path):
        root_path = Path(root)
        res += (root_path / file for file in files)
    return res


def resolve_local_path(path: str) -> List[Path]:
    if "*" in path:
        return _resolve_glob(path)
    if path.endswith("/"):
        return _resolve_dir(path)
    return [Path(path)]



import os
import platform
import subprocess

import setuptools

PROJECT_SRC_DIR = os.path.abspath(os.path.join(os.getcwd(), "../../.."))
CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))
PROTO_DIR = os.path.abspath(os.path.join(CURRENT_DIR, "clickzetta/proto/source"))
PROTO_OUT_DIR = os.path.abspath(os.path.join(CURRENT_DIR, "clickzetta/proto/generated"))

os.makedirs(PROTO_OUT_DIR, exist_ok=True)

for source_file in os.listdir(PROTO_DIR):
    subprocess.call(
        'python -m grpc_tools.protoc -I . --python_out=' + PROTO_OUT_DIR + ' --grpc_python_out=' + PROTO_OUT_DIR
        + ' --proto_path =' + PROTO_DIR + ' '
        + os.path.abspath(os.path.join(PROTO_DIR, source_file)), shell=True)

for generated_file in os.listdir(PROTO_OUT_DIR):
    if platform.system() == "Darwin":
        subprocess.call("sed -i '' 's/^import /from . import /' " + os.path.abspath(
            os.path.join(PROTO_OUT_DIR, generated_file)), shell=True)
    elif platform.system() == "Linux":
        subprocess.call("sed -i 's/^import /from . import /' " + os.path.abspath(
            os.path.join(PROTO_OUT_DIR, generated_file)), shell=True)

# Package metadata.

name = "clickzetta-connector"
description = "clickzetta python connector"

# Should be one of:
# 'Development Status :: 3 - Alpha'
# 'Development Status :: 4 - Beta'
# 'Development Status :: 5 - Production/Stable'
release_status = "Development Status :: 3 - Alpha"
dependencies = [
    "proto-plus >= 1.22.0, <2.0.0dev",
    "packaging >= 14.3, <24.0.0dev",
    "protobuf>=3.19.5,<5.0.0dev,!=3.20.0,!=3.20.1,!=4.21.0,!=4.21.1,!=4.21.2,!=4.21.3,!=4.21.4,!=4.21.5",
    "python-dateutil >= 2.7.2, <3.0dev",
    "requests >= 2.21.0, < 3.0.0dev",
    "pyarrow >= 10.0.1, <11.0.0",
    "numpy == 1.26.4",
    "sqlalchemy >= 1.4.0, <2.0.0",
    #"ossfs >= 2023.1.0",
    "cz-ossfs >= 0.0.2",
    "cos-python-sdk-v5 >= 1.9.25",
    "pandas >=1.5.3",
    "s3fs[boto3] == 2023.5.0",
    "boto3 == 1.28.17",
    "google-cloud-storage <= 2.17.0",
    "gcsfs <= 2024.6.0",
]

# Setup boilerplate below this line.

package_root = os.path.abspath(os.path.dirname(__file__))

version = {}
with open(os.path.join(package_root, "clickzetta/version.py")) as fp:
    exec(fp.read(), version)
version = version["__version__"]

packages = ['clickzetta', 'clickzetta.dbapi', 'clickzetta.bulkload', 'clickzetta.proto.generated']

setuptools.setup(
    name=name,
    version=version,
    description=description,
    url='https://www.zettadecision.com/',
    author="mocun",
    author_email="hanmiao.li@clickzetta.com",
    platforms="Posix; MacOS X;",
    packages=packages,
    install_requires=dependencies,
    python_requires=">=3.7",
    include_package_data=True,
    zip_safe=False,
)



from clickzetta.dbapi import exceptions
import functools


def raise_on_closed(
        exc_msg, exc_class=exceptions.ProgrammingError, closed_attr_name="_closed"
):

    def _raise_on_closed(method):

        def with_closed_check(self, *args, **kwargs):
            if getattr(self, closed_attr_name):
                raise exc_class(exc_msg)
            return method(self, *args, **kwargs)

        functools.update_wrapper(with_closed_check, method)
        return with_closed_check

    def decorate_public_methods(klass):
        for name in dir(klass):
            if name.startswith("_") and name != "__iter__":
                continue

            member = getattr(klass, name)
            if not callable(member):
                continue

            if isinstance(klass.__dict__[name], (staticmethod, classmethod)):
                continue

            member = _raise_on_closed(member)
            setattr(klass, name, member)

        return klass

    return decorate_public_methods


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: file_system.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11\x66ile_system.proto\x12\x08\x63z.proto*A\n\x0e\x46ileSystemType\x12\t\n\x05LOCAL\x10\x00\x12\x07\n\x03OSS\x10\x01\x12\t\n\x05\x43\x41\x43HE\x10\x02\x12\x07\n\x03RPC\x10\x03\x12\x07\n\x03\x43OS\x10\x04\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'file_system_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_FILESYSTEMTYPE']._serialized_start=31
  _globals['_FILESYSTEMTYPE']._serialized_end=96
# @@protoc_insertion_point(module_scope)
